import pika.channel
from easy_amqp import EasyAMQP
from easy_amqp.models import Message, ExchangeType
from pika.connection import ConnectionParameters

import pika

credentials = pika.PlainCredentials('test', 'test')
connection_params = ConnectionParameters("localhost", credentials=credentials)

rabbit = EasyAMQP(connection_parameters=connection_params)



@rabbit.declare_queue("test_queue") # will declare a queue named "test_queue"
@rabbit.declare_exchange("test_exchange", exchange_type=ExchangeType.direct) # will declare an exchange named "test_exchange" of type direct
@rabbit.bind(exchange="test_exchange", queue="test_queue", routing_key="test_routing_key") # exchange will send messages to the queue with the routing key "test_routing_key"
@rabbit.listen("test_queue", message_type=str) # will listen to the queue "test_queue" and consume messages as strings
@rabbit.batch()
def consume(message: Message, channel: pika.channel.Channel):
    print(message.body) # will be a List[str] due to the batch decorator


rabbit.run()


import pika
import pika.channel

import pika.spec
from typing import Callable, List, Dict, Optional, Type, TypeVar, Any
from dataclasses import dataclass, field
from enum import StrEnum

T = TypeVar('T')
F = TypeVar('F', bound=Callable[..., Any])

@dataclass
class Retry:
    retries: int
    delay: float

@dataclass
class DeadLetterQueue:
    x_dead_letter_exchange: str
    x_dead_letter_routing_key: str
    x_max_length: Optional[int] = None
    x_message_ttl:  Optional[int] = None

    def to_args(self) -> Dict[str, Any]:
        args: Dict[str, Any] = {}
        if self.x_max_length:
            args['x-max-length'] = int(self.x_max_length)
        if self.x_message_ttl:
            args['x-message-ttl'] = int(self.x_message_ttl)
        args['x-dead-letter-exchange'] = self.x_dead_letter_exchange
        args['x-dead-letter-routing-key'] = self.x_dead_letter_routing_key
        return args

@dataclass
class Queue:
    name: str
    passive: bool
    durable: bool
    exclusive: bool
    auto_delete: bool
    arguments: Optional[Dict[str, Any]]
    dead_letter_queue: Optional[DeadLetterQueue]
    on_queue_declared: Optional[Callable[[pika.channel.Channel, 'Queue'], None]] = None


class ExchangeType(StrEnum):
    direct = 'direct'
    fanout = 'fanout'
    headers = 'headers'
    topic = 'topic'


@dataclass
class Exchange:
    name: str
    type: ExchangeType
    durable: bool
    passive: bool
    auto_delete: bool
    internal: bool
    arguments: Optional[Dict[str, Any]]
    on_exchange_declared: Optional[Callable[[pika.channel.Channel, 'Exchange'], None]] = None


@dataclass
class Binding:
    exchange: str
    routing_key: str
    queue: str
    arguments: Optional[Dict[str, Any]]
    on_binding_declared: Optional[Callable[[pika.channel.Channel, 'Binding'], None]] = None


@dataclass
class Message:
    body: T
    method: pika.spec.Basic.Deliver
    props: pika.spec.BasicProperties


@dataclass
class Batch:
    batch_time: float
    scheduled: bool
    items: List[Message] = field(default_factory=list)


@dataclass
class Listener:
    queue: str
    func: Callable[..., None] 
    message_type: Type[T]
    auto_ack: bool
    exclusive: bool
    consumer_tag: Optional[str]
    arguments: Optional[Dict[str, Any]]
    parser: Optional[Callable[[bytes], T]]
    on_channel_open: Optional[Callable[[pika.channel.Channel], None]]


@dataclass
class Prefetch:
    size: int = 0
    count: int = 0
    global_qos: bool = False


@dataclass
class Handler:
    listener: Listener
    batch: Optional[Batch] = None
    prefetch: Optional[Prefetch] = None


import pika.channel
from easy_amqp import EasyAMQP
from easy_amqp.models import Message
from pika.connection import ConnectionParameters

import pika

"""
This example shows how to use a custom parser function to convert the message body from bytes to a custom object."""

credentials = pika.PlainCredentials('test', 'test')
connection_params = ConnectionParameters("localhost", credentials=credentials)

rabbit = EasyAMQP(connection_parameters=connection_params)

class Data:
    def __init__(self, value: str):
        self.value = value



def custom_parser(body: bytes) -> Data:
    """
    Custom parser function to convert bytes to str.
    This is useful if you want to handle the message body in a specific way.
    """
    return Data(body.decode('utf-8'))

@rabbit.listen("test", message_type=Data, custom_parser=custom_parser)
def consume(message: Message, channel: pika.channel.Channel):
    print(message.body.value)


rabbit.run()


import pika
import pytest
import requests
from testcontainers.rabbitmq import RabbitMqContainer
from easy_amqp.EasyAMQP import EasyAMQP

from typing import Dict, Any, Iterator



@pytest.fixture
def rabbitmq_container() -> Iterator[Dict[str, Any]]:
    """
    Pytest fixture to set up and tear down a RabbitMQ container for the entire test session.
    Yields a dictionary containing host, client_port, and api_port.
    """
    port_client: int = 5672
    port_api: int = 15672
    with RabbitMqContainer("rabbitmq:3.12-management").with_exposed_ports(port_client, port_api) as container:
        host: str = container.get_container_host_ip()
        client_port: int = container.get_exposed_port(port_client)
        api_port: int = container.get_exposed_port(port_api)
        yield {"host": host, "client_port": client_port, "api_port": api_port}

@pytest.fixture
def easy_amqp(rabbitmq_container: Dict[str, Any]) -> Iterator[EasyAMQP]:
    """
    Pytest fixture to provide an EasyAMQP instance, depending on the RabbitMQ container.
    A new instance is provided for each test function.
    """
    credentials = pika.PlainCredentials("guest", "guest")
    connection_params = pika.ConnectionParameters(
        host=rabbitmq_container["host"],
        credentials=credentials,
        port=rabbitmq_container["client_port"]
    )
    easy_rabbit: EasyAMQP = EasyAMQP(connection_params)
    yield easy_rabbit
    # Ensure easy_rabbit is stopped after each test using this fixture
    easy_rabbit.stop()


def get_exchange_properties(host: str, port_api: int, exchange_name: str) -> Dict[str, Any]:
    """
    Retrieves exchange properties from the RabbitMQ management API.
    """
    url = f"http://{host}:{port_api}/api/exchanges/%2F/{exchange_name}"
    response = requests.get(url, auth=("guest", "guest"))
    response.raise_for_status()
    return response.json()

def get_queue_properties(host: str, port_api: int, queue_name: str) -> Dict[str, Any]:
    """
    Retrieves queue properties from the RabbitMQ management API.
    """
    url = f"http://{host}:{port_api}/api/queues/%2F/{queue_name}"
    response = requests.get(url, auth=("guest", "guest"))
    response.raise_for_status()
    return response.json()




import pika.channel
from easy_amqp import EasyAMQP
from easy_amqp.models import Message, Retry

import pika

"""
This example shows how to use multiple connection parameters and how to set a retry policy if the connection fails.
"""

rabbit = EasyAMQP(connection_parameters=[
    pika.ConnectionParameters('rabbitmq1'),
    pika.ConnectionParameters('rabbitmq2'),
], retry=Retry(3, 5.0))

@rabbit.listen("test", message_type=str)
def consume(message: Message, channel: pika.channel.Channel):
    print(message.body) # will be a str due to the message_type parameter


rabbit.run()


import pika.channel
import pika.exceptions
import pika
import pytest
from ..conftest import get_queue_properties
from easy_amqp.EasyAMQP import EasyAMQP
from easy_amqp.models import Queue, Message, DeadLetterQueue
from typing import Dict, Any, Union



def test_add_queue(
    rabbitmq_container: Dict[str, Any],
    easy_amqp: EasyAMQP
) -> None:
    """
    Tests the creation of a queue using EasyAMQP.
    """
    result: Dict[str, Any] = {}

    def on_queue_declared(_: pika.channel.Channel, q: Queue) -> None:
        result['queue'] = get_queue_properties(rabbitmq_container["host"], rabbitmq_container["api_port"], q.name)
        easy_amqp.stop()

    queue = Queue(
        name="test_queue",
        passive=False,
        durable=True,
        exclusive=True,
        auto_delete=True,
        arguments=None,
        dead_letter_queue=None,
        on_queue_declared=on_queue_declared
    )
    easy_amqp.add_queue(queue)
    easy_amqp.run()

    if 'queue' not in result:
        pytest.fail("Queue declaration callback was not called")

    assert result['queue']['name'] == queue.name
    # assert result['queue']['durable'] is True #ToDo check why this is not working
    assert result['queue']['auto_delete'] == queue.auto_delete
    assert result['queue']['arguments'] == {}
    assert result['queue']['exclusive'] is queue.exclusive


def test_add_queue_with_decorator(
    rabbitmq_container: Dict[str, Any],
    easy_amqp: EasyAMQP
) -> None:
    """
    Tests the creation of a queue using the EasyAMQP decorator.
    """
    result: Dict[str, Any] = {}
    queue_name: str = "test_queue"

    def on_queue_declared(_: pika.channel.Channel, q: Queue) -> None:
        result['queue'] = get_queue_properties(rabbitmq_container["host"], rabbitmq_container["api_port"], q.name)
        easy_amqp.stop()

    @easy_amqp.declare_queue(queue_name, on_queue_declared=on_queue_declared)
    def consuming(msg: Message, _: pika.channel.Channel) -> None: # type: ignore
        pass

    easy_amqp.run()

    if 'queue' not in result:
        pytest.fail("Queue declaration callback was not called")
    assert result['queue']['name'] == queue_name


def test_add_queue_passive_mode_queue_does_not_exists(
    rabbitmq_container: Dict[str, Any],
) -> None:
    """
    Tests adding a queue in passive mode when the queue does not exist.
    Expects a ChannelClosedByBroker exception.
    """
    result: Dict[str, Any] = {'ok': False}

    def channel_closed(_: pika.channel.Channel, exception: pika.exceptions.ChannelClosed) -> None:
        result['ok'] = isinstance(exception, pika.exceptions.ChannelClosedByBroker) and "NOT_FOUND" in str(exception)
        easy_rabbit.stop() # Stop this specific instance

    # Should not go to this callback because passive=True will cause the broker to close the channel automatically
    def on_queue_declared(_: pika.channel.Channel, q: Queue) -> None:
        easy_rabbit.stop() # Stop this specific instance

    # Create a new EasyAMQP instance for this specific test case,
    # as its behavior (on_channel_closed) is unique.
    credentials = pika.PlainCredentials("guest", "guest")
    connection_params = pika.ConnectionParameters(
        host=rabbitmq_container["host"],
        credentials=credentials,
        port=rabbitmq_container["client_port"]
    )
    easy_rabbit = EasyAMQP(connection_params, on_channel_closed=channel_closed)

    easy_rabbit.add_queue(Queue(
        name="test_queue",
        passive=True,
        durable=True,
        exclusive=True,
        auto_delete=True,
        arguments=None,
        dead_letter_queue=None,
        on_queue_declared=on_queue_declared
    ))
    easy_rabbit.run()

    assert result['ok'], "Expected channel to be closed with a 'no queue' error, but it was not."



def setup_rabbit_with_queue(rabbitmq_container: Dict[str, Any], queue: str) -> None:
    connection = pika.BlockingConnection(
        pika.ConnectionParameters(
            host=rabbitmq_container["host"],
            credentials=pika.PlainCredentials("guest", "guest"),
            port=rabbitmq_container["client_port"]
        )
    )
    channel = connection.channel()
    channel.queue_declare(queue=queue) # type: ignore
    connection.close()


def test_add_queue_passive_mode_queue_exists(
    rabbitmq_container: Dict[str, Any],
) -> None:
    """
    Tests adding a queue in passive mode when the queue already exists.
    """
    result: Dict[str, Any] = {'ok': False}

    # Declare the queue manually before EasyAMQP attempts to declare it passively
    setup_rabbit_with_queue(rabbitmq_container, "test_queue")
    def channel_closed(_: pika.channel.Channel, reason: Union[pika.exceptions.ChannelClosed, str]) -> None:
        result['ok'] = False
        easy_rabbit.stop() # Stop this specific instance

    def on_queue_declared(_: pika.channel.Channel, q: Queue) -> None:
        queue_properties: Dict[str, Any] = get_queue_properties(rabbitmq_container["host"], rabbitmq_container["api_port"], q.name)
        result['ok'] = queue_properties['name'] == "test_queue"
        easy_rabbit.stop() # Stop this specific instance

    # Create a new EasyAMQP instance for this specific test case,
    # as its behavior (on_channel_closed) is unique.
    credentials = pika.PlainCredentials("guest", "guest")
    connection_params = pika.ConnectionParameters(
        host=rabbitmq_container["host"],
        credentials=credentials,
        port=rabbitmq_container["client_port"]
    )
    easy_rabbit: EasyAMQP = EasyAMQP(connection_params, on_channel_closed=channel_closed)

    easy_rabbit.add_queue(Queue(
        name="test_queue",
        passive=True,
        durable=True,
        exclusive=True,
        auto_delete=True,
        arguments=None,
        dead_letter_queue=None,
        on_queue_declared=on_queue_declared
    ))
    easy_rabbit.run()
    assert result['ok']


def test_add_queue_with_dead_letter(
    rabbitmq_container: Dict[str, Any],
    easy_amqp: EasyAMQP
) -> None:
    """
    Tests the creation of a queue with dead-letter queue arguments.
    """
    result: Dict[str, Any] = {}

    def on_queue_declared(_: pika.channel.Channel, q: Queue) -> None:
        result['queue'] = get_queue_properties(rabbitmq_container["host"], rabbitmq_container["api_port"], q.name)
        easy_amqp.stop()

    dlq = DeadLetterQueue(
        x_dead_letter_exchange="dlx_exchange",
        x_dead_letter_routing_key="dlx_routing_key",
        x_max_length=100,
        x_message_ttl=60000
    )
    easy_amqp.add_queue(Queue(
        name="test_queue",
        passive=False,
        durable=False,
        exclusive=False,
        auto_delete=False,
        arguments=None,
        dead_letter_queue=dlq,
        on_queue_declared=on_queue_declared
    ))
    easy_amqp.run()

    if 'queue' not in result:
        pytest.fail("Queue declaration callback was not called")

    assert result['queue']['arguments'] == {
        'x-dead-letter-exchange': dlq.x_dead_letter_exchange,
        'x-dead-letter-routing-key': dlq.x_dead_letter_routing_key,
        'x-max-length': dlq.x_max_length,
        'x-message-ttl': dlq.x_message_ttl
    }


def test_add_queue_with_dlq_and_arguments(
    rabbitmq_container: Dict[str, Any],
    easy_amqp: EasyAMQP
) -> None:
    """
    Tests the creation of a queue with dead-letter queue arguments and additional custom arguments.
    """
    result: Dict[str, Any] = {}

    def on_queue_declared(_: pika.channel.Channel, q: Queue) -> None:
        result['queue'] = get_queue_properties(rabbitmq_container["host"], rabbitmq_container["api_port"], q.name)
        easy_amqp.stop()

    dlq = DeadLetterQueue(
        x_dead_letter_exchange="dlx_exchange",
        x_dead_letter_routing_key="dlx_routing_key",
        x_max_length=100,
        x_message_ttl=60000
    )
    easy_amqp.add_queue(Queue(
        name="test_queue",
        passive=False,
        durable=False,
        exclusive=False,
        auto_delete=False,
        arguments={'x-queue-mode': 'lazy'},
        dead_letter_queue=dlq,
        on_queue_declared=on_queue_declared
    ))
    easy_amqp.run()

    if 'queue' not in result:
        pytest.fail("Queue declaration callback was not called")

    assert result['queue']['arguments'] == {
        'x-dead-letter-exchange': dlq.x_dead_letter_exchange,
        'x-dead-letter-routing-key': dlq.x_dead_letter_routing_key,
        'x-max-length': dlq.x_max_length,
        'x-message-ttl': dlq.x_message_ttl,
        'x-queue-mode': 'lazy'
    }


import pika
import json

credentials = pika.PlainCredentials('test', 'test')
connection = pika.BlockingConnection(pika.ConnectionParameters('localhost', credentials=credentials))
channel = connection.channel()

test_value = "test"
for counter in range(10):
    channel.basic_publish(exchange='',
                        routing_key=test_value,
                        body=f"{test_value} {counter}")

connection.close()


import pika.connection
from .TopologyManager import TopologyManager
import pika.channel
from typing import Callable, Any
from .logger import LOGGER
import pika
from functools import partial
from .models import Queue, Exchange, Binding


class Declarator:
    def __init__(self, connection: pika.connection.Connection, channel: pika.channel.Channel, topology_manager: TopologyManager, on_finish: Callable[[pika.connection.Connection, pika.channel.Channel], None]) -> None:
        self._connection = connection
        self._channel = channel
        self._topology_manager = topology_manager
        self._on_finish = on_finish
        declarations =  len(self._topology_manager.queues()) + len(self._topology_manager.exchanges()) + len(self._topology_manager.bindings())
        self._remaining = declarations
        self._count_declarations = declarations


    def declare(self):
        if self._remaining == 0:
            LOGGER.debug("No declarations to perform, moving on..")
            self._on_finish(self._connection, self._channel)
            return
        
        self._declare_queues()
        self._declare_exchanges()
        self._apply_bindings()

    def _task_done(self, callable_fn: Callable[[], None], _: Any = None) -> None:
        self._remaining -= 1
        LOGGER.debug(f"declaration task done, remaining: {self._remaining} of {self._count_declarations}")
        callable_fn()
        
        if self._remaining == 0:
            self._on_finish(self._connection, self._channel)

    def _declare_queues(self):
        for queue in self._topology_manager.queues():
            LOGGER.debug(f"declare queue: {queue.name}")
            args = (queue.arguments or {}).copy()
            if queue.dead_letter_queue:
                args.update(queue.dead_letter_queue.to_args())

            def make_callback_fn(queue_snapshot: Queue) -> Callable[[], None]:
                def callback() -> None:
                    LOGGER.debug(f"queue: {queue_snapshot.name} is declared")
                    if queue_snapshot.on_queue_declared:
                        queue_snapshot.on_queue_declared(self._channel, queue_snapshot)
                return callback
            callback_fn = make_callback_fn(queue)
            self._channel.queue_declare(
                queue=queue.name,
                passive=queue.passive,
                durable=queue.durable,
                exclusive=queue.exclusive,
                auto_delete=queue.auto_delete,
                arguments=args,
                callback=partial(self._task_done, callback_fn)
            )



    def _declare_exchanges(self):
        for exchange in self._topology_manager.exchanges():
            LOGGER.debug(f"declare exchange: {exchange.name}")


            def make_callback_fn(exchange_snapshot: Exchange) -> Callable[[], None]:
                def callback() -> None:
                    LOGGER.debug(f"{exchange_snapshot.name} is declared")
                    if exchange_snapshot.on_exchange_declared:
                        exchange_snapshot.on_exchange_declared(self._channel, exchange_snapshot)
                return callback
            
            callback_fn = make_callback_fn(exchange)

            self._channel.exchange_declare(
                exchange=exchange.name,
                exchange_type=exchange.type.value,
                passive=exchange.passive,
                durable=exchange.durable,
                auto_delete=exchange.auto_delete,
                internal=exchange.internal,
                arguments=exchange.arguments,
                callback=partial(self._task_done, callback_fn)
            )
    
    def _apply_bindings(self) -> None:
        for binding_config in self._topology_manager.bindings():
            LOGGER.debug(f"bind queue: {binding_config.queue} an exchange: {binding_config.exchange} with routingkey: {binding_config.routing_key}")
            def make_callback_fn(binding_snapshot: Binding) -> Callable[[], None]:
                def callback() -> None:
                    LOGGER.debug( f"bind queue: {binding_snapshot.queue} an exchange: {binding_snapshot.exchange} with routingkey: {binding_snapshot.routing_key} binded")
                return callback
            callback_fn = make_callback_fn(binding_config)

            self._channel.queue_bind(
                exchange=binding_config.exchange, queue=binding_config.queue,
                routing_key=binding_config.routing_key, arguments=binding_config.arguments,
                callback=partial(self._task_done, callback_fn)
            )


import pika.channel
import pytest
from ..conftest import get_exchange_properties
from easy_amqp.EasyAMQP import EasyAMQP
from easy_amqp.models import Exchange, ExchangeType, Message
from typing import Dict, Any

def test_add_exchange(
    rabbitmq_container: Dict[str, Any],
    easy_amqp: EasyAMQP
) -> None:
    """
    Tests the creation of an exchange using EasyAMQP.
    """
    result: Dict[str, Any] = {}

    def on_exchange_declared(_: pika.channel.Channel, e: Exchange) -> None:
        result['exchange'] = get_exchange_properties(rabbitmq_container["host"], rabbitmq_container["api_port"], e.name)
        easy_amqp.stop()

    exchange: Exchange = Exchange(
        name="test_exchange",
        type=ExchangeType.fanout,
        durable=True,
        passive=False,
        internal=False,
        auto_delete=False,
        arguments=None,
        on_exchange_declared=on_exchange_declared
    )
    easy_amqp.add_exchange(exchange)
    easy_amqp.run()

    if 'exchange' not in result:
        pytest.fail("Exchange declaration callback was not called")

    assert result['exchange']['name'] == exchange.name
    assert result['exchange']['durable'] is True #ToDo check why this is not working
    assert result['exchange']['auto_delete'] == exchange.auto_delete
    assert result['exchange']['arguments'] == {}


def test_add_exchange_with_decorator(
    rabbitmq_container: Dict[str, Any],
    easy_amqp: EasyAMQP
) -> None:
    """
    Tests the creation of an exchange using the EasyAMQP decorator.
    """
    result: Dict[str, Any] = {}
    exchange_name: str = "test_exchange"

    def on_exchange_declared(_: pika.channel.Channel, e: Exchange) -> None:
        result['exchange'] = get_exchange_properties(rabbitmq_container["host"], rabbitmq_container["api_port"], e.name)
        easy_amqp.stop()

    @easy_amqp.declare_exchange(exchange=exchange_name, exchange_type=ExchangeType.fanout, on_exchange_declared=on_exchange_declared)
    def consuming(msg: Message, channel: pika.channel.Channel) -> None: # type: ignore
        pass

    easy_amqp.run()

    if 'exchange' not in result:
        pytest.fail("Exchange declaration callback was not called")
    assert result['exchange']['name'] == exchange_name


from easy_amqp import EasyAMQP
from pika.connection import ConnectionParameters
import pika
from easy_amqp.models import Queue, Exchange, ExchangeType


credentials = pika.PlainCredentials('test', 'test')
connection_params = ConnectionParameters("localhost", credentials=credentials)

rabbit = EasyAMQP(connection_parameters=connection_params)

"""
This example shows how to use the EasyAMQP library to create a queue and an exchange without using decorators, 
and then run the RabbitMQ server.
"""


rabbit.add_queue(Queue(
    name="test_queue",
    passive=True,
    durable=True,
    exclusive=True,
    auto_delete=True,
    arguments=None,
    dead_letter_queue=None,
))

exchange: Exchange = Exchange(
        name="test_exchange",
        type=ExchangeType.fanout,
        durable=True,
        passive=False,
        internal=False,
        auto_delete=False,
        arguments=None
    )
rabbit.add_exchange(exchange)

rabbit.run()


import pika.channel
import pika.connection
import pika.exceptions
from easy_amqp import EasyAMQP
from easy_amqp.models import Message
from pika.connection import ConnectionParameters
from typing import Union
import pika


def on_connection_open(connection: pika.connection.Connection):
    print("Connection opened successfully")

def on_connection_error(connection: pika.connection.Connection, exception: Union[str, Exception]):
    print(f"Connection error: {exception}")

def on_channel_closed(channel: pika.channel.Channel, exception: pika.exceptions.ChannelClosed):
    print(f"Channel: {channel} closed unexpectedly reason: {exception}")

credentials = pika.PlainCredentials('test', 'test')
connection_params = ConnectionParameters("localhost", credentials=credentials)

rabbit = EasyAMQP(connection_parameters=connection_params, on_connection_open=on_connection_open, on_connection_error=on_connection_error, on_channel_closed=on_channel_closed)



@rabbit.listen("test", message_type=str)
def consume(message: Message, channel: pika.channel.Channel):
    print(message.body) # will be a str due to the message_type parameter


rabbit.run()


import logging

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(level=logging.DEBUG)

handler = logging.StreamHandler()
formatter = logging.Formatter(
    fmt="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S")
handler.setFormatter(formatter)

LOGGER.addHandler(handler)


import pika.channel
import pika
import pytest
from easy_amqp.EasyAMQP import EasyAMQP
from easy_amqp.models import Message
from typing import List, Dict, Any


def test_consuming(
    rabbitmq_container: Dict[str, Any],
    easy_amqp: EasyAMQP
) -> None:
    """
    Tests a basic message consumption scenario.
    """
    result: Dict[str, Any] = {}
    queue_name: str = "test_queue"
    exchange_name: str = "test_exchange"
    message: str = "Test message"

    # Setup RabbitMQ with a blocking connection to publish a message
    connection: pika.BlockingConnection = pika.BlockingConnection(
        pika.ConnectionParameters(
            host=rabbitmq_container["host"],
            credentials=pika.PlainCredentials("guest", "guest"),
            port=rabbitmq_container["client_port"]
        )
    )
    channel = connection.channel()
    channel.queue_declare(queue=queue_name) # type: ignore
    channel.exchange_declare(exchange=exchange_name, exchange_type='direct') # type: ignore
    channel.queue_bind(queue=queue_name, exchange=exchange_name, routing_key="test_routing_key") # type: ignore
    channel.basic_publish(exchange=exchange_name, routing_key="test_routing_key", body=message.encode('utf-8'))
    connection.close()

    @easy_amqp.listen(queue_name, message_type=str)
    def consuming(msg: Message, _: pika.channel.Channel) -> None: # type: ignore
        result['message'] = msg.body
        easy_amqp.stop()

    easy_amqp.run()

    if 'message' not in result:
        pytest.fail("Consumption callback was not called")
    assert result['message'] == message


def test_consuming_batch(
    rabbitmq_container: Dict[str, Any],
    easy_amqp: EasyAMQP
) -> None:
    """
    Tests consuming multiple messages in a batch.
    """
    result: Dict[str, Any] = {}
    queue_name = "test_queue"
    exchange_name = "test_exchange"
    message_prefix = "Test message"

    # Setup RabbitMQ with a blocking connection to publish multiple messages
    connection = pika.BlockingConnection(
        pika.ConnectionParameters(
            host=rabbitmq_container["host"],
            credentials=pika.PlainCredentials("guest", "guest"),
            port=rabbitmq_container["client_port"]
        )
    )
    channel = connection.channel() # type: ignore
    channel.queue_declare(queue=queue_name) # type: ignore
    channel.exchange_declare(exchange=exchange_name, exchange_type='direct') # type: ignore
    channel.queue_bind(queue=queue_name, exchange=exchange_name, routing_key="test_routing_key") # type: ignore
    for counter in range(5):
        channel.basic_publish(exchange=exchange_name, routing_key="test_routing_key", body=f"{message_prefix} {counter}".encode('utf-8'))
    connection.close()

    @easy_amqp.listen(queue_name, message_type=str)
    @easy_amqp.batch()
    def consuming(msgs: List[Message], _: pika.channel.Channel) -> None: # type: ignore
        result['messages'] = msgs
        easy_amqp.stop()

    easy_amqp.run()

    if 'messages' not in result:
        pytest.fail("Consumption callback was not called")

    assert isinstance(result['messages'], list)
    assert len(result['messages']) == 5 # type: ignore
    for i, msg in enumerate(result['messages']): # type: ignore
        assert msg.body == f"{message_prefix} {i}" # type: ignore


import pika
import pika.channel
import pika.spec
from typing import List
from .models import Message, Handler
from . import parser
from .logger import LOGGER

def start_consuming(handler: Handler, channel: pika.channel.Channel):
    listener = handler.listener
    LOGGER.debug(f"Setting up consumer for queue {listener.queue} on channel {channel.channel_number}")
    if listener.on_channel_open:
        listener.on_channel_open(channel)
    if handler.prefetch:
        pf = handler.prefetch
        channel.basic_qos(
            prefetch_count=pf.count,
            prefetch_size=pf.size,
            global_qos=pf.global_qos
        )
    channel.basic_consume(
        queue=listener.queue,
        on_message_callback=lambda ch, method, props, body: dispatch(handler, ch, method, props, body),
        auto_ack=listener.auto_ack,
        exclusive=listener.exclusive,
        consumer_tag=listener.consumer_tag,
        arguments=listener.arguments
    )

def dispatch(handler: Handler, channel: pika.channel.Channel,
                method: pika.spec.Basic.Deliver, props: pika.spec.BasicProperties, body: bytes) -> None:
    if handler.listener.parser:
        parsed_body = handler.listener.parser(body)
    else:
        parsed_body = parser.default_parser(body, handler.listener.message_type)
    msg = Message(body=parsed_body, method=method, props=props)

    if handler.batch:
        batch = handler.batch
        batch.items.append(msg)
        if not batch.scheduled:
            batch.scheduled = True
            channel.connection.ioloop.call_later( # type: ignore
                batch.batch_time,
                lambda: _deliver_batch(handler, channel)
            )
    else:
        _callback(handler, msg, channel)

def _deliver_batch(handler: Handler, channel: pika.channel.Channel):
    batch = handler.batch
    assert batch is not None
    items, batch.items = batch.items, []
    batch.scheduled = False
    _callback(handler, items, channel)



def _callback(handler: Handler, message: Message | List[Message], channel: pika.channel.Channel):
    handler.listener.func(message, channel)


from typing import Type
from .models import T

from .logger import LOGGER
import json


def default_parser(body: bytes, message_type: Type[T]) -> T:
    if message_type is bytes:
        return body # type: ignore
    elif message_type is str:
        return body.decode('utf-8') # type: ignore
    elif message_type in (int, float, bool):
        parsed_json = json.loads(body)
        return message_type(parsed_json) # type: ignore
    else:
        try:
            data = json.loads(body)
            return message_type(**data)
        except (json.JSONDecodeError, TypeError) as e:
            try: 
                data = json.loads(body)
                return message_type(data) # type: ignore
            except (json.JSONDecodeError, TypeError) as exp:
                LOGGER.debug(f"Unable to parse message to typ {message_type}: {e}")
                raise ValueError(f"Unable to parse message to typ {message_type}: {e}") from exp


import pika
import threading
import pika.channel
from typing import Callable, List, Optional,Union
import time

import pika.exceptions
from .TopologyManager import TopologyManager
from .models import Retry
import pika.connection
from . import consuming
from .logger import LOGGER
from .Declarator import Declarator
from functools import partial



class ConnectionManager:
    def __init__(self,
                 params: Union[pika.ConnectionParameters, List[pika.ConnectionParameters]],
                 topology_manager: TopologyManager,
                 retry: Optional[Retry] = None,
                 on_open: Optional[Callable[[pika.connection.Connection], None]] = None,
                 on_error: Optional[Callable[[pika.connection.Connection, Union[str, Exception]], None]] = None,
                 on_channel_closed: Optional[Callable[[pika.channel.Channel, pika.exceptions.ChannelClosed], None]] = None):
        self._params = params
        self._topology_manager = topology_manager
        self._retry = retry
        self._custom_on_open = on_open
        self._custom_on_error = on_error
        self._custom_on_channel_closed = on_channel_closed
        self._connection: Optional[pika.SelectConnection] = None
        self._ioloop_thread: Optional[threading.Thread] = None
        self._last_used_index = -1

        self._connection_closed = False

    def _choose_params(self) -> pika.ConnectionParameters:
        if isinstance(self._params, list):
            if len(self._params) == 1:
                return self._params[0]
            if self._last_used_index != -1:
                self._last_used_index = (self._last_used_index + 1) % len(self._params)
            else:
                self._last_used_index = 0
            return self._params[self._last_used_index]
        return self._params

    def _connect(self) -> pika.SelectConnection:
        return pika.SelectConnection(
            parameters=self._choose_params(),
            on_open_callback=self._on_open,
            on_open_error_callback=self._on_error_callback
        )

    def _on_open(self, conn: pika.connection.Connection):
        LOGGER.debug("Connection opened")
       
        if self._custom_on_open:
            self._custom_on_open(conn)
        conn.channel(on_open_callback=partial(self._on_channel_open_declare, conn))
      

    def _on_declaration_finish(self, conn: pika.connection.Connection, declaration_channel: pika.channel.Channel) -> None:
        if self._connection_closed:
            LOGGER.debug("connection is already closed, cannot proceed with finish declaration")
            return
        LOGGER.debug("All declarations finished, starting consumers")
        declaration_channel.close()  # type: ignore
        for handler in self._topology_manager.handlers():
            channel = conn.channel(on_open_callback=lambda ch, h=handler: consuming.start_consuming(h, ch))
            channel.add_on_close_callback(self._on_channel_closed)    # type: ignore   


    def _on_channel_closed(self, channel: pika.channel.Channel, exception: pika.exceptions.ChannelClosed):
        if self._connection_closed:
            LOGGER.debug(f"cannot close channel because, connection is already closed, reason for channel close: {str(exception)}")
            return

        LOGGER.debug(f"Channel {channel.channel_number}, reason: {str(exception)}")
        if self._custom_on_channel_closed:
            self._custom_on_channel_closed(channel, exception)
        
        # channel is not closed by code
        if not isinstance(exception, pika.exceptions.ChannelClosedByClient):
            LOGGER.warning(f"Channel {channel.channel_number} not closed by broker: {str(exception)}")
            self.close()



    def _on_channel_open_declare(self, conn: pika.connection.Connection, channel: pika.channel.Channel) -> None:
        try:
            LOGGER.debug(f"channel open for declaration channel number: {channel.channel_number}")
            channel.add_on_close_callback(self._on_channel_closed)  # type: ignore
            declarator = Declarator(conn, channel, self._topology_manager, self._on_declaration_finish)
            declarator.declare()
        except:
            LOGGER.exception("Error during channel declaration")
            self.close()
            raise

    def _on_error_callback(self, conn: pika.connection.Connection, err: Union[str, Exception]):
        LOGGER.error(f"Connection error: {err}")
        if self._custom_on_error:
            self._custom_on_error(conn, err)
        raise ConnectionError("Unable to establish connection")

    def start(self):
        attempts = 0
        while True:
            try:
                self._connection = self._connect()
                self._connection.ioloop.start() # type: ignore       
                break
            except Exception as e:
                if not self._retry:
                    self.close()
                    raise e
                
                attempts += 1
                LOGGER.info(f"Reconnect attempt #{attempts} waiting: {self._retry.delay} second(s)")
                if self._retry.retries == -1:
                    time.sleep(self._retry.delay)
                    continue
                if attempts >= self._retry.retries:
                    LOGGER.error("Max retries reached, exiting.")
                    self.close()
                    raise e
               
                time.sleep(self._retry.delay)

    def run_in_thread(self) -> threading.Thread:
        self.ioloop_thread = threading.Thread(target=self.start, daemon=True)
        self.ioloop_thread.start()
        return self.ioloop_thread
    
    def close(self) -> None:
        if self._connection_closed:
            LOGGER.debug("Connection already closed")
            return
        if self._connection and not self._connection.is_closed:
            self._connection.close()
        if self._connection and self._connection.ioloop: # type: ignore
            self._connection.ioloop.stop() # type: ignore
        if hasattr(self, "_ioloop_thread") and self._ioloop_thread:
             self._ioloop_thread.join()
        self._connection_closed = True
        LOGGER.debug("connection closed and IOLoop stopped")

    def get_connection(self) -> pika.SelectConnection:
        if self._connection is None:
            raise ValueError("no connection is open")
        return self._connection


import pika
import threading
import pika.channel
from typing import Callable, List, Dict, Optional, Type, Any, Union

import pika.exceptions
from .ConnectionManager import ConnectionManager
from .TopologyManager import TopologyManager, get_func_identifier
import pika.connection
from .models import T, F, Retry, DeadLetterQueue, Queue, ExchangeType, Exchange, Binding, Batch, Listener, Prefetch


class EasyAMQP:
    def __init__(self, 
                 connection_parameters: pika.ConnectionParameters | List[pika.ConnectionParameters], 
                 retry: Optional[Retry] = None,
                 on_connection_open: Optional[Callable[[pika.connection.Connection], None]]= None,
                 on_connection_error: Optional[Callable[[pika.connection.Connection, Union[str, Exception]], None]] = None,
                 on_channel_closed: Optional[Callable[[pika.channel.Channel, pika.exceptions.ChannelClosed], None]] = None
        ) -> None:        
        self._topology_manager = TopologyManager()
        self._connection_manager = ConnectionManager(
            params=connection_parameters,
            topology_manager=self._topology_manager,
            retry=retry,
            on_open=on_connection_open,
            on_error=on_connection_error,
            on_channel_closed=on_channel_closed
        )


    def declare_queue(self, queue: str, passive: bool = False, durable: bool = False, exclusive: bool = False, auto_delete: bool = False, arguments: Optional[Dict[str, Any]] = None, on_queue_declared: Optional[Callable[[pika.channel.Channel, 'Queue'], None]] = None) -> Callable[[F], F]:
        def decorator(func: F) -> F:
            self._topology_manager.add_queue(get_func_identifier(func), Queue(
                name=queue,
                passive=passive,
                durable=durable,
                exclusive=exclusive,
                auto_delete=auto_delete,
                arguments=arguments,
                dead_letter_queue=None,
                on_queue_declared=on_queue_declared
            ))
            return func
        return decorator

    def declare_exchange(self, exchange: str, exchange_type: ExchangeType = ExchangeType.direct, passive: bool = False, durable: bool = False, auto_delete: bool = False, internal: bool = False, arguments: Optional[Dict[str, Any]] = None, on_exchange_declared: Optional[Callable[[pika.channel.Channel, 'Exchange'], None]] = None) -> Callable[[F], F]:
        def decorator(func: F) -> F:
            self._topology_manager.add_exchange(get_func_identifier(func), Exchange(
                name=exchange,
                type=exchange_type,
                passive=passive,
                auto_delete=auto_delete,
                durable=durable,
                arguments=arguments,
                internal=internal,
                on_exchange_declared=on_exchange_declared
            ))
            return func
        return decorator

    def bind(self, exchange:str, queue: str, routing_key: str, arguments: Optional[Dict[str, Any]] = None, on_binding_declared: Optional[Callable[[pika.channel.Channel, 'Binding'], None]] = None) -> Callable[[F], F]:
        def decorator(func: F) -> F:
            self._topology_manager.add_binding(Binding(
                exchange=exchange,
                queue=queue,
                routing_key=routing_key,
                arguments=arguments,
                on_binding_declared=on_binding_declared
            ))
            return func
        return decorator

    def prefetch(self, prefetch_count: int = 0, prefetch_size: int = 0, global_qos: bool = False) -> Callable[[F], F]:
        def decorator(func: F) -> F:
            self._topology_manager.add_prefetch(get_func_identifier(func), Prefetch(
                    count=prefetch_count,
                    size=prefetch_size,
                    global_qos=global_qos
            ))
            return func
        return decorator
    
    def dead_letter(self, x_dead_letter_exchange: str, x_dead_letter_routing_key: str,  x_max_length: Optional[int] = None, x_message_ttl:  Optional[int] = None) -> Callable[[F], F]:
        def decorator(func: F) -> F:
            self._topology_manager.add_dead_letter(get_func_identifier(func), DeadLetterQueue(
                x_dead_letter_exchange=x_dead_letter_exchange,
                x_dead_letter_routing_key=x_dead_letter_routing_key,
                x_max_length=x_max_length,
                x_message_ttl=x_message_ttl
            ))

            return func
        return decorator

    def batch(self, batch_time: float = 0.2) -> Callable[[F], F]:
        def decorator(func: F) -> F:
            self._topology_manager.add_batch_consumer(get_func_identifier(func), Batch(
                batch_time=batch_time,
                scheduled=False,
                items=[]
            ))
            return func
        return decorator
    
    def listen_batch(self, queue: str, message_type: Type[T], batch_time: float = 0.2, auto_ack: bool = True, exclusive: bool = False, consumer_tag: Optional[str] = None, arguments: Optional[Dict[str, Any]] = None, custom_parser: Optional[Callable[[bytes], T]] = None, on_channel_open: Optional[Callable[[pika.channel.Channel], None]] = None) -> Callable[[F], F]:
        def decorator(func: F) -> F:
            self._topology_manager.add_batch_consumer(get_func_identifier(func), Batch(
                batch_time=batch_time,
                scheduled=False,
                items=[]
            ))
            self._topology_manager.add_listener(get_func_identifier(func), Listener(
                    func=func,
                    queue=queue,
                    message_type=message_type,
                    auto_ack=auto_ack,
                    exclusive=exclusive,
                    consumer_tag=consumer_tag,
                    arguments=arguments,
                    parser=custom_parser,
                    on_channel_open=on_channel_open
            ))
            return func
        return decorator

    def listen(self, queue: str, message_type: Type[T], auto_ack: bool = True, exclusive: bool = False, consumer_tag: Optional[str] = None, arguments: Optional[Dict[str, Any]] = None, custom_parser: Optional[Callable[[bytes], T]] = None, on_channel_open: Optional[Callable[[pika.channel.Channel], None]] = None) -> Callable[[F], F]:
        def decorator(func: F) -> F: 
            self._topology_manager.add_listener(get_func_identifier(func), Listener(
                    func=func,
                    queue=queue,
                    message_type=message_type,
                    auto_ack=auto_ack,
                    exclusive=exclusive,
                    consumer_tag=consumer_tag,
                    arguments=arguments,
                    parser=custom_parser,
                    on_channel_open=on_channel_open
            ))
    
            return func
        return decorator


    def add_queue(self, queue: Queue) -> None:
        self._topology_manager.add_queue(queue.name, queue)
    
    def add_exchange(self, exchange: Exchange) -> None:
        self._topology_manager.add_exchange(exchange.name, exchange)

    def add_binding(self, binding: Binding) -> None:
        self._topology_manager.add_binding(binding)

    def run_in_thread(self) -> threading.Thread:
        return self._connection_manager.run_in_thread()

    def run(self) -> None:
        self._connection_manager.start()

    def stop(self) -> None:
        self._connection_manager.close() 




from typing import Callable, List, Dict, Any
from .models import DeadLetterQueue, Queue, Exchange, Binding, Batch, Listener, Prefetch, Handler
from .logger import LOGGER

def get_func_identifier(func: Callable[..., Any]) -> str:
    return f"{func.__module__}.{func.__qualname__}"

class TopologyManager:
    def __init__(self) -> None:
        self._listener: Dict[str, Listener] = {}
        self._prefetch: Dict[str, Prefetch] = {}
        self._batch: Dict[str, Batch] = {}
        self._queues_to_declare: Dict[str, Queue] = {}
        self._exchanges_to_declare: Dict[str, Exchange] = {}
        self._dead_letter_queues: Dict[str, DeadLetterQueue] = {}
        self._bindings: List[Binding] = []


    def add_queue(self, caller_name: str, queue: Queue) -> None:
        for func_name, queue_to_declare in self._queues_to_declare.items():
            if queue.name == queue_to_declare.name:
                LOGGER.warning(f"queue: {queue.name} is already declared in {func_name}, skipping")
                raise ValueError(f"queue: {queue} is already declared in {func_name}")
        if caller_name in self._dead_letter_queues:
            queue.dead_letter_queue = self._dead_letter_queues[caller_name]
        self._queues_to_declare[caller_name] = queue
        
    
    def add_exchange(self, caller_name: str, exchange: Exchange) -> None:
        for func_name, exchange_to_declare in self._exchanges_to_declare.items():
            if exchange.name == exchange_to_declare.name:
                LOGGER.warning(f"Exchange: {exchange} is already declared in {func_name}")
                raise ValueError(f"Exchange: {exchange} is already declared in {func_name}")
        self._exchanges_to_declare[caller_name] = exchange

    def add_binding(self, binding: Binding) -> None:
        self._bindings.append(binding)

    def add_prefetch(self, caller_name: str, prefetch: Prefetch) -> None:
        self._prefetch[caller_name] = prefetch

    def add_dead_letter(self, caller_name: str, dlq: DeadLetterQueue) -> None:
        self._dead_letter_queues[caller_name] = dlq

    def add_batch_consumer(self, caller_name: str, batch: Batch) -> None:
        self._batch[caller_name] = batch

    def add_listener(self, caller_name: str, listener: Listener) -> None:
        self._listener[caller_name] = listener

    def handlers(self) -> List[Handler]:
        return [Handler(
            listener=listener,
            prefetch=self._prefetch[key] if key in self._prefetch else None,
            batch=self._batch[key] if key in self._batch else None
        )        
        for key, listener in self._listener.items()]
    
    def queues(self) -> List[Queue]:
        for key, queue in self._queues_to_declare.items():
            if key in self._dead_letter_queues:
                queue.dead_letter_queue = self._dead_letter_queues[key]
                
        return list(self._queues_to_declare.values())
    
    def bindings(self) -> List[Binding]:
        return self._bindings

    def exchanges(self) -> List[Exchange]:
        return list(self._exchanges_to_declare.values())
    

   


import pika.channel
from easy_amqp import EasyAMQP
from easy_amqp.models import Message
from pika.connection import ConnectionParameters

import pika

"""
This example shows how to use the listen to the queue "test" to consume messages.
The listen decorator will consume messages one by one and pass them to the consumer function."""

credentials = pika.PlainCredentials('test', 'test')
connection_params = ConnectionParameters("localhost", credentials=credentials)

rabbit = EasyAMQP(connection_parameters=connection_params)

@rabbit.listen("test", message_type=str)
def consume(message: Message, channel: pika.channel.Channel):
    print(message.body) # will be a str due to the message_type parameter


rabbit.run()